@@ -1,14 +1,14 @@
 ;;; cape.el --- Completion At Point Extensions -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021-2023 Free Software Foundation, Inc.
+;; Copyright (C) 2021-2024 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
-;; Version: 1.1
+;; Version: 1.2
 ;; Package-Requires: ((emacs "27.1") (compat "29.1.4.4"))
 ;; Homepage: https://github.com/minad/cape
-;; Keywords: abbrev, convenience, matching, completion, wp
+;; Keywords: abbrev, convenience, matching, completion, text
 
 ;; This file is part of GNU Emacs.
 
@@ -902,12 +902,20 @@ The functions `cape-wrap-super' and `cape-capf-super' are experimental."
                                 (sort (or (completion-metadata-get md 'display-sort-function)
                                           #'identity))
                                 (cands (funcall sort (all-completions str table pr))))
+                           ;; Handle duplicates with a hash table.
                            (cl-loop
-                            for cand in-ref cands do
-                            (if (eq (gethash cand ht t) t)
-                                (puthash cand plist ht)
-                              (setf cand (propertize cand #'cape-capf-super
-                                                     (cons cand plist)))))
+                            for cand in-ref cands
+                            for dup = (gethash cand ht t) do
+                            (cond
+                             ((eq dup t)
+                              ;; Candidate does not yet exist.
+                              (puthash cand plist ht))
+                             ((not (equal dup plist))
+                              ;; Duplicate candidate. Candidate plist is
+                              ;; different, therefore disambiguate the
+                              ;; candidates.
+                              (setf cand (propertize cand 'cape-capf-super
+                                                     (cons cand plist))))))
                            (push cands candidates)))
                 (setq cand-ht ht)
                 (apply #'nconc (nreverse candidates))))
@@ -925,7 +933,7 @@ The functions `cape-wrap-super' and `cape-capf-super' are experimental."
         ,@(mapcan
            (lambda (prop)
              (list prop (lambda (cand &rest args)
-                          (let ((ref (get-text-property 0 #'cape-capf-super cand)))
+                          (let ((ref (get-text-property 0 'cape-capf-super cand)))
                             (when-let ((fun (plist-get
                                              (or (cdr ref)
                                                  (and cand-ht (gethash cand cand-ht)))
@@ -998,7 +1006,8 @@ completion table is refreshed on every input change."
                (input (buffer-substring-no-properties beg end)))
           (lambda (str pred action)
             (let ((new-input (buffer-substring-no-properties beg end)))
-              (unless (or (cape--separator-p new-input)
+              (unless (or (not (eq action t))
+                          (cape--separator-p new-input)
                           (funcall valid input new-input))
                 (pcase
                     ;; Reset in case `all-completions' is used inside CAPF
@@ -1148,43 +1157,36 @@ This function can be used as an advice around an existing Capf."
     (`(,beg ,end ,table . ,plist)
      `(,beg ,end ,(cape--accept-all-table table) . ,plist))))
 
-(defmacro cape--capf-wrapper (wrapper)
-  "Create a Capf transformer from WRAPPER."
-  `(defun ,(intern (format "cape-capf-%s" wrapper)) (&rest args)
-     (lambda () (apply #',(intern (format "cape-wrap-%s" wrapper)) args))))
-
 ;;;###autoload (autoload 'cape-capf-accept-all "cape")
-(cape--capf-wrapper accept-all)
 ;;;###autoload (autoload 'cape-capf-buster "cape")
-(cape--capf-wrapper buster)
 ;;;###autoload (autoload 'cape-capf-case-fold "cape")
-(cape--capf-wrapper case-fold)
 ;;;###autoload (autoload 'cape-capf-debug "cape")
-(cape--capf-wrapper debug)
 ;;;###autoload (autoload 'cape-capf-inside-comment "cape")
-(cape--capf-wrapper inside-comment)
 ;;;###autoload (autoload 'cape-capf-inside-faces "cape")
-(cape--capf-wrapper inside-faces)
 ;;;###autoload (autoload 'cape-capf-inside-string "cape")
-(cape--capf-wrapper inside-string)
-;;;###autoload (autoload 'cape-capf-super "cape")
-(cape--capf-wrapper super)
-;;;###autoload (autoload 'cape-capf-noninterruptible "cape")
-(cape--capf-wrapper noninterruptible)
 ;;;###autoload (autoload 'cape-capf-nonexclusive "cape")
-(cape--capf-wrapper nonexclusive)
+;;;###autoload (autoload 'cape-capf-noninterruptible "cape")
 ;;;###autoload (autoload 'cape-capf-passthrough "cape")
-(cape--capf-wrapper passthrough)
 ;;;###autoload (autoload 'cape-capf-predicate "cape")
-(cape--capf-wrapper predicate)
 ;;;###autoload (autoload 'cape-capf-prefix-length "cape")
-(cape--capf-wrapper prefix-length)
 ;;;###autoload (autoload 'cape-capf-properties "cape")
-(cape--capf-wrapper properties)
 ;;;###autoload (autoload 'cape-capf-purify "cape")
-(cape--capf-wrapper purify)
 ;;;###autoload (autoload 'cape-capf-silent "cape")
-(cape--capf-wrapper silent)
+;;;###autoload (autoload 'cape-capf-super "cape")
+
+(dolist (wrapper (list #'cape-wrap-accept-all #'cape-wrap-buster
+                       #'cape-wrap-case-fold #'cape-wrap-debug
+                       #'cape-wrap-inside-comment #'cape-wrap-inside-faces
+                       #'cape-wrap-inside-string #'cape-wrap-nonexclusive
+                       #'cape-wrap-noninterruptible #'cape-wrap-passthrough
+                       #'cape-wrap-predicate #'cape-wrap-prefix-length
+                       #'cape-wrap-properties #'cape-wrap-purify
+                       #'cape-wrap-silent #'cape-wrap-super))
+  (let ((name (string-remove-prefix "cape-wrap-" (symbol-name wrapper))))
+    (defalias (intern (format "cape-capf-%s" name))
+      (lambda (capf &rest args) (lambda () (apply wrapper capf args)))
+      (format "Create a %s Capf from CAPF.
+The Capf calls `%s' with CAPF and ARGS as arguments." name wrapper))))
 
 (provide 'cape)
 ;;; cape.el ends here
